##
# Copyright 2014, Aurel Wildfellner.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#

from threading import Lock, Thread
import time

import spaceapi


class EventType:
    WASOPENED = "wasopened"
    WASCLOSED = "wasclosed"
    # manual locking
    WASLOCKED = "waslocked"
    WASUNLOCKED = "wasunlocked"
    # lock commands
    DOLOCK = "dolock"
    DOUNLOCK = "dounlock"
    DOLOCKAFTERME = "dolockafterme"
    ERROR = "error"
    INIT = "init"


class EventSource:
    WEBINTERFACE = "webinterface"
    SHELL = "shell" 
    INTERNAL = "internal"
    DOOR = "door"


class LockSystemEvent:

    def __init__(self, source, etype):
        self.source = source
        self.etype = etype

    def toString(self):
        return "[" + self.etype + "] from (" + self.source + ")"



class EventQueue:

    def __init__(self):
        self._lock = Lock()
        self.events = []
        pass


    def flushEvents(self):
        with self._lock:
            self.events = []


    def enqueueEvent(self, event):
        with self._lock:
            self.events.append(event)


    def getNextEvent(self):
        """ get next event - returns None if nothing in the queue"""
        with self._lock:
            if (len(self.events) == 0):
                return None
            else:
                return self.events.pop(0)
           

    def getNextEventBlocking(self):
        """ get next event - blocking """
        # this is not a good way to handle locks (use RIIA)
        while True:
            with self._lock:
                if (len(self.events) != 0):
                    e = self.events.pop(0)
                    return e
            time.sleep(0.1)
            



class LockSystemFSM(object):

    def __init__(self, door, eventQueue):
        self.door = door
        self.eventQueue = eventQueue
        self._stopThread = False


    def start(self):
        self._thread = Thread(target=self.run)
        self._stopThread = False
        self._thread.start()
        self._threadRunning = True


    def stop(self):
        self._stopThread = True
        while (self._threadRunning):
            time.sleep(0.01)


    def run(self):
        currentState = self.state_init
        ev = LockSystemEvent(EventSource.INTERNAL, EventType.INIT)

        while not (self._stopThread):
            ### do the state change
            print("Processing event: ", ev.toString())
            print("State Transition: ", currentState.__name__, "--> ", end="")
            lastState = currentState
            currentState = currentState(ev)
            print(currentState.__name__)
            #print("Waiting for next event.... ", end="")

            ### state change detected
            if (currentState != lastState):
                # detect spaceapi lock event
                if (currentState == self.state_closedUnlocked):
                    spaceapi.setDoorLockStatus(False)
                if (currentState == self.state_closedLocked):
                    spaceapi.setDoorLockStatus(True)

            ### get next event
            while (not self._stopThread):
                ev = self.eventQueue.getNextEvent()
                if (not ev):
                    time.sleep(0.1)
                else:
                    break
            #print("got it.")

        self._threadRunning = False


    ##### states ####

    def state_init(self, event):
        """check the current state of the system and initialize accordingly"""
        if (self.door.isOpen()):
            # if the door is open and unlock, reset to unlocked
            if (self.door.isLocked()):
                self.door.unlock()
            return self.state_open

        else:
            if (self.door.isLocked()):
                return self.state_closedLocked
            else:
                return self.state_closedUnlocked


    def state_open(self, event):
        """open, locked or unlocked, wait for events"""
        if (event.etype == EventType.WASCLOSED):
            # not that the door can't be closed when locked
            # so the next state must be closedUnlocked!!!
            return self.state_closedUnlocked
        if (event.etype == EventType.DOLOCKAFTERME):
            return self.state_doLockAfterMeOpen
        if (event.etype == EventType.DOUNLOCK):
            self.door.unlock()
            return self.state_open

        return self.state_open


    def state_openLocked(self, event):
        """this may happen when the door is smashed in or
           when there are problems with the sensor."""
        if (event.etype == EventType.WASCLOSED):
            return self.state_closedLocked
        if (event.etype == EventType.DOUNLOCK):
            self.door.unlock()
            return self.state_openLocked
        if (event.etype == EventType.WASUNLOCKED):
            return self.state_open

        return self.state_openLocked


    def state_doLockAfterMeOpen(self, event):
        """open; next time it is closed lock automatically"""
        if (event.etype == EventType.WASCLOSED):
            # some delay here to make sure the door is really shut
            time.sleep(1)
            self.door.lock()
            return self.state_closedUnlocked
        if (event.etype == EventType.DOUNLOCK):
            self.door.unlock()
            return self.state_doLockAfterMeOpen
            
        return self.state_doLockAfterMeOpen


    def state_doLockAfterMeClosed(self, event):
        """closed, next time open+close lock automatically"""
        if (event.etype == EventType.WASOPENED):
            return self.state_doLockAfterMeOpen
        if (event.etype == EventType.WASLOCKED):
            return self.state_closedLocked
        if (event.etype == EventType.DOLOCK):
            self.door.lock()
            return self.state_doLockAfterMeClosed

        return self.state_doLockAfterMeClosed


    def state_closedUnlocked(self, event):
        """the door is closed and unlocked, can be opened or locked in several ways"""
        if (event.etype == EventType.WASLOCKED):
            return self.state_closedLocked
        if (event.etype == EventType.DOLOCK):
            self.door.lock()
            return self.state_closedUnlocked
        if (event.etype == EventType.DOLOCKAFTERME):
            return self.state_doLockAfterMeClosed
        if (event.etype == EventType.WASOPENED):
            return self.state_open
        # also send unlock command in this state
        if (event.etype == EventType.DOUNLOCK):
            self.door.unlock()
            return self.state_closedUnlocked

        return self.state_closedUnlocked


    def state_closedLocked(self, event):
        """closed and locked, wait for unlocking"""
        # possible break in!!!1
        if (event.etype == EventType.WASOPENED):
            return self.state_openLocked
        if (event.etype == EventType.WASUNLOCKED):
            return self.state_closedUnlocked
        if (event.etype == EventType.DOUNLOCK):
            self.door.unlock()
            return self.state_closedLocked

        return self.state_closedLocked


    def state_error(self, event):
        """something has gone wrong, reset"""
        # generate some sort of event
        ev = LockSystemEvent(EventSource.INTERNAL, EventType.ERROR)
        self.eventQueue.flushEvents()
        self.eventQueue.enqueueEvent(ev)
        return self.state_init



