##
# Copyright 2013, Aurel Wildfellner.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#

import serial
from abc import ABCMeta, abstractmethod

from corefsm import *



class Door(object):

    __metaclass__ = ABCMeta


    @abstractmethod
    def isOpen(self):
        pass

    @abstractmethod
    def isLocked(self):
        pass

    @abstractmethod
    def lock(self):
        pass

    @abstractmethod
    def unlock(self):
        pass



class MockupDoor(Door):

    def __init__(self, eventQueue):
        self._isOpen = False
        self._isLocked = True
        self.eventQueue = eventQueue
        pass


    def isOpen(self):
        return self._isOpen 


    def isLocked(self):
        return self._isLocked


    def lock(self):
        if not (self._isOpen):
            if not (self._isLocked):
                self._isLocked = True
                ev = LockSystemEvent(EventSource.DOOR, EventType.WASLOCKED)
                self.eventQueue.enqueueEvent(ev)


    def unlock(self):
        if (self._isLocked):
            self._isLocked = False
            ev = LockSystemEvent(EventSource.DOOR, EventType.WASUNLOCKED)
            self.eventQueue.enqueueEvent(ev)


    def setOpenState(self, state):
        """Only for testing. Should detect a switch."""
        if (self._isOpen != state):
            self._isOpen = state
            if (self._isOpen):
                ev = LockSystemEvent(EventSource.DOOR, EventType.WASOPENED)
            else:
                ev = LockSystemEvent(EventSource.DOOR, EventType.WASCLOSED)
            self.eventQueue.enqueueEvent(ev)


    def setLockedState(self, state):
        """Only for testing. Should detect lock ADC."""
        if (self._isLocked != state):
            self._isLocked = state
            if (self._isLocked):
                ev = LockSystemEvent(EventSource.DOOR, EventType.WASLOCKED)
            else:
                ev = LockSystemEvent(EventSource.DOOR, EventType.WASUNLOCKED)
            self.eventQueue.enqueueEvent(ev)




class ArduinoDoor(Door):


    def __init__(self, eventQueue):
        self._isOpen = False
        self._isLocked = True
        self._serial = None
        self.eventQueue = eventQueue

        # threading
        self._thread = None
        self._stopThread = False

        self.isConnected = False


    def isOpen(self):
        self.sendCommand("GETSTATE")
        # a delay actually works! timeout -> oldstate
        time.sleep(0.3)
        return self._isOpen


    def isLocked(self):
        self.sendCommand("GETSTATE")
        time.sleep(0.3)
        return self._isLocked


    def lock(self):
        self.sendCommand("DOLOCK")


    def unlock(self):
        self.sendCommand("DOUNLOCK")


    #### serial connection to arduino ###

    def connectArduino(self, device="/dev/ttyACM0", baudrate=9600):
        if self.isConnected:
            return True

        try:
            self._serial = serial.Serial(device, baudrate)
            self.startThread()
        except:
            return False
        # update open and locked info
        self.isConnected = self._serial.isOpen()
        return self.isConnected


    def disconnectArduino(self):
        self.stopThread()
        if self._serial:
            self._serial.close()
        self.isConnected = False
    

    def sendCommand(self, cmd):
        """ Sends a command string to the arduino. """
        if self._serial and self._serial.isOpen():
            cmd += ";"
            try:
                self._serial.write(bytearray(cmd, "ASCII"))
            except:
                pass


    def startThread(self):
        if self._thread:
            return
        self._thread = Thread(target=self.runThread)
        self._thread.start()


    def stopThread(self):
        thread = self._thread
        if thread:
            self._stopThread = True
            thread.join()
            self._thread = None
            self._stopThread = False
    

    def runThread(self):
        while not self._stopThread and self._serial.isOpen():
            try:
                bl = self._serial.readline()
            except:
                self.isConnected = False
                break
            line = bl.decode('unicode_escape')

            ### process events ###
            if "WASUNLOCKED" in line:
                ev = LockSystemEvent(EventSource.DOOR, EventType.WASUNLOCKED)
                self.eventQueue.enqueueEvent(ev)
                self._isLocked = False

            elif "WASLOCKED" in line:
                ev = LockSystemEvent(EventSource.DOOR, EventType.WASLOCKED)
                self.eventQueue.enqueueEvent(ev)
                self._isLocked = True

            elif "WASOPENED" in line:
                ev = LockSystemEvent(EventSource.DOOR, EventType.WASOPENED)
                self.eventQueue.enqueueEvent(ev)
                self._isOpen = True

            elif "WASCLOSED" in line:
                ev = LockSystemEvent(EventSource.DOOR, EventType.WASCLOSED)
                self.eventQueue.enqueueEvent(ev)
                self._isOpen = False
           
            # response to GETSTATE
            #FIXME bad protocol
            elif "STATE" in line:
                states = [ str.strip(s) for s in line.split(":")[1].split(",") ]
                if "LOCKED" in states:
                    self._isLocked = True
                if "UNLOCKED" in states:
                    self._isLocked = False
                if "OPEN" in states:
                    self._isOpen = True
                if "CLOSED" in states:
                    self._isOpen = False

        self._thread = None


